#include "ClockedBufferArray.h"

#include <assert.h>

#include "Messages_m.h"
#include "Util.h"

namespace YARC {

/*
 * This is the base class for the row and column buffer arrays. It contains
 * an array of input queue pairs ( one for each VC ). The "current" input queue
 * is varied in a round robin wormhole fashion ( ie. a packet is never cut in the
 * middle but after one has been completely sent a new queue is chosen ). On each
 * clock a new flit from the "current" input queue is forwarded if possible. The
 * canSend* hooks are called to determine if it's possible to send. The congestion
 * state of the input queue pair array is also managed in this base class and when
 * a congestion happens a sideband message is sent over the congestion gate.
 */

void ClockedBufferArray::initialize() {
    in = gate("in");
    out = gate("out");
    congestion = gate("congestion");
    
    clockStart = par("bufferClockStart");
    clockPeriod = par("bufferClockPeriod");
    bufferThreshold = par("bufferThreshold");
    bufferDepth = par("bufferDepth");
    peerCount = par("peerCount");
    
    queues = new cQueue[peerCount][2];
    
    currentInputVC = new char[peerCount];
    memset(currentInputVC, -1, sizeof(char) * peerCount);
    
    inputCongested = new char[peerCount][2];
    memset(inputCongested, 0, sizeof(char[2]) * peerCount);
    
    currentOutputSourceQueue = -1;
    currentOutputVC = -1;
    
    lastOutputSourceQueue = -1;
    lastOutputVC = 1;
    
    cMessage *msg = new cMessage();
    msg->setSchedulingPriority(32000);
    scheduleAt(clockStart, msg);
    
    WATCH(clockStart);
    WATCH(clockPeriod);
    WATCH(bufferThreshold);
    WATCH(currentOutputSourceQueue);
    WATCH(currentOutputVC);
    WATCH(lastOutputSourceQueue);
    WATCH(lastOutputVC);
    WATCH(peerCount);
    for (int i=0; i < peerCount; i++) {
        const int BUFF_SIZE = 100;
        char *buff = newBuff(BUFF_SIZE);      
        
        snprintf(buff, BUFF_SIZE, "queues[%d][0]", i);
        createWatch_cObject(buff, queues[i][0]);
        
        snprintf(buff, BUFF_SIZE, "queues[%d][1]", i);
        createWatch_cObject(buff, queues[i][1]);
        
        snprintf(buff, BUFF_SIZE, "currentInputVC[%d]", i);
        watchFlag(buff, currentInputVC[i]);
        
        snprintf(buff, BUFF_SIZE, "inputCongested[%d][0]", i);
        watchFlag(buff, inputCongested[i][0]);
        
        snprintf(buff, BUFF_SIZE, "inputCongested[%d][1]", i);
        watchFlag(buff, inputCongested[i][1]);
    }
}

void ClockedBufferArray::updateSideband() {
    for (int peer=0; peer < peerCount; peer++) {
        bool mustUpdate = false;
        char currentSideband[2] = {0, 0};
        for (int vc=0; vc < 2; vc++) {
            if (queues[peer][vc].getLength() >= bufferThreshold) {
                currentSideband[vc] = 1;
            }
            if (currentSideband[vc] != inputCongested[peer][vc]) {
                mustUpdate = true;
            }
        }

        if (mustUpdate) {
            Sideband *sbd = new Throttle();

            addLongPar(sbd, UNICAST_DST_ATTR, peer);

            copySideband(currentSideband, sbd);
            copySideband(currentSideband, inputCongested[peer]);

            send(sbd, congestion);
        }

        if (bufferDepth != -1) {
            assert(queues[peer][0].getLength() <= bufferDepth);
            assert(queues[peer][1].getLength() <= bufferDepth);
        }
    }
}

void ClockedBufferArray::roundRobinStep(int &peerId, int &vcId) {
    if (vcId == 0) {
        vcId = 1;
    } else {
        peerId = (peerId + 1) % peerCount;
        vcId = 0;
    } 
}

void ClockedBufferArray::output() {
    if (currentOutputSourceQueue == -1) {
        for (int i=0; i < peerCount * 2; i++) {
            roundRobinStep(lastOutputSourceQueue, lastOutputVC);
            
            cQueue &queue = queues[lastOutputSourceQueue][lastOutputVC];
            if (!queue.isEmpty()) {
                HeadFlit *headFlit = check_and_cast<HeadFlit*>(queue.front());
                if (canSendHeadFlit(headFlit)) {
                    currentOutputSourceQueue = lastOutputSourceQueue;
                    currentOutputVC = lastOutputVC;
                    break;
                }
            }
        }
    } 

    if (currentOutputSourceQueue != -1) {                
        cQueue &queue = queues[currentOutputSourceQueue][currentOutputVC];

        if (!queue.isEmpty()) {
            Flit *flit = check_and_cast<Flit*>(queue.front());            
            HeadFlit *headFlit = dynamic_cast<HeadFlit*>(flit);
            TailFlit *tailFlit = dynamic_cast<TailFlit*>(flit);            
            bool canSend = false;            
            if (headFlit != NULL) {
                canSend = true;
            } else {    
                if (tailFlit != NULL) {
                    canSend = canSendTailFlit(tailFlit);
                    if (canSend) {
                        currentOutputSourceQueue = -1;
                        currentOutputVC = -1;
                    }
                } else {
                    canSend = canSendFlit(flit);                
                }
            }

            if (canSend) {
                queue.pop();
                send(flit, out);
            }
        }
    }
}

void ClockedBufferArray::handleMessage(cMessage *msg) {
    if (msg->getArrivalGate() == in) {
        Flit *flit = check_and_cast<Flit*>(msg);
        
        int srcPeer = xtrLongPar(flit, UNICAST_SRC_ATTR);
        assert(srcPeer >= 0 && srcPeer < peerCount);
        
        HeadFlit *headFlit = dynamic_cast<HeadFlit*>(msg);
        if (headFlit != NULL) {
            assert(currentInputVC[srcPeer] == -1);            
            currentInputVC[srcPeer] = headFlit->getVC();
        }
        
        assert(currentInputVC[srcPeer] == 0 || currentInputVC[srcPeer] == 1);               
        int srcPeerCurVC = currentInputVC[srcPeer];        
        queues[srcPeer][srcPeerCurVC].insert(flit);
        
        TailFlit *tailFlit = dynamic_cast<TailFlit*>(msg);
        if (tailFlit != NULL) {
            currentInputVC[srcPeer] = -1;
        }
    } else if (msg->isSelfMessage()) {
        delete msg;
        
        output();       
        updateSideband();
        
        cMessage *msg = new cMessage();
        msg->setSchedulingPriority(32000);
        simtime_t nextClock = simTime();
        nextClock += clockPeriod;
        scheduleAt(nextClock, msg);
    }
}

int ClockedBufferArray::getPeerCount() {
    return peerCount;
}

int ClockedBufferArray::getCurrentOutputVC() {
    return currentOutputVC;
}

int ClockedBufferArray::getCurrentOutputSourceQueue() {
    return currentOutputSourceQueue;
}

}
