#include "Host.h"

#include <assert.h>

using namespace std;

namespace YARC {

/*
 * This module serves to create traffic in the network. Concrete simulations
 * are expected to inherit it and override the sendNewPacket and receiveNewPacket
 * methods. 
 * 
 * On every clock cycle during which the host isn't sending flits sendNewPacket is called
 * and the concrete host implementation can choose to either send a new packet 
 * ( by returning a pair containing a HeadFlit instance and packet length ) or do nothing by 
 * returning the (NULL, -1) pair. Any traffic pattern can be implemented by selectively ignoring
 * or using opportunities to send. 
 * 
 * The arguments to sendNewPacket contain the current information about the input buffer state on 
 * the other side of the link. Trying to send into a know-to-be-congested buffer will abort the 
 * simulation.
 * 
 * If sendNewPacket returns a non (NULL, -1) pair, one flit is sent on every clock. Until the packet
 * has been fully sent sendNewPacket is not called again. 
 * 
 * Once all flits of an incoming packet have been receive receiveNewPacket is called. This hook
 * can be used in subclasses for logging and statistics.  
 */

Define_Module(Host);

void Host::initialize() {
    linkRx = gate("link$i");
    linkTx = gate("link$o");
    
    clockStart = par("hostClockStart");
    clockPeriod = par("hostClockPeriod");
    
    curRxHead = NULL;
    curRxLength = -1;
    rxDone = false;
    
    curTxHead = NULL;
    curTxLength = -1;
    sideband[0] = 0;
    sideband[1] = 0;
    
    cMessage *msg = new cMessage();
    msg->setSchedulingPriority(32000);
    scheduleAt(clockStart, msg);
    
    WATCH(clockStart);
    WATCH(clockPeriod);
    WATCH(curRxLength);
    WATCH(rxDone);
    WATCH(curTxLength);
    WATCH_FLAG(sideband[0]);
    WATCH_FLAG(sideband[1]);
}

pair<HeadFlit*, int> Host::sendNewPacket(const char (&sideband)[2]) {
    // By default do nothing
    // pair<HeadFlit*, int>(NULL, -1) and generally any pair with length -1 means 
    // "do not send anything this clock"
    return pair<HeadFlit*, int>(NULL, -1);   
}

void Host::receiveNewPacket(int length, HeadFlit *head) {
    // By default do nothing
    return; 
}

void Host::clock() {
    simtime_t now = simTime();

    if (rxDone) {
    	curRxHead->addPar(RECEIVE_TIME).setDoubleValue(now.dbl());
    	receiveNewPacket(curRxLength, curRxHead);
    	delete curRxHead->removeObject(RECEIVE_TIME);
    	delete curRxHead->removeObject(SEND_TIME);
    	
        rxDone = false;
        curRxLength = -1;
        delete curRxHead;
        curRxHead = NULL;
    }        

    if (curTxLength == -1 /* ie. tx done */) {
        pair<HeadFlit*, int> newPacket = sendNewPacket(sideband);
        
        if (!(newPacket.first == NULL && newPacket.second == -1)) {
            assert(newPacket.first->getVC() == 0 || newPacket.first->getVC() == 1);
            assert(sideband[ newPacket.first->getVC() ] == 0);

            newPacket.first->addPar(SEND_TIME).setDoubleValue(now.dbl());
        }
            
        curTxHead = newPacket.first;
        curTxLength = newPacket.second;
    } 

    assert(curTxLength > 0 || curTxLength == -1); 
    if (curTxLength > 0) {
        if (curTxHead != NULL) {
            assert(curTxLength >= 2);

            send(curTxHead, linkTx);
            curTxHead = NULL;
        } else {
            if (curTxLength == 1) {
                send(new TailFlit(), linkTx);
            } else {
                send(new Flit(), linkTx);
            }
        }
        
        curTxLength--;
        if (curTxLength == 0) {
            curTxLength = -1;
        }
    }
    
    cMessage *msg = new cMessage();
    msg->setSchedulingPriority(32000);
    scheduleAt(now + clockPeriod, msg);
}

void Host::handleMessage(cMessage *msg) {
    if (msg->isSelfMessage()) {
        clock();
        delete msg;
    } else {
        assert(msg->getArrivalGate() == linkRx);
        
        Flit *flit = check_and_cast<Flit*>(msg);
        
        copySideband(flit, sideband);
        
        IdleFlit *idle = dynamic_cast<IdleFlit*>(msg);
        if (idle != NULL) {
            delete idle;
            
            return;
        }
        
        if (curRxHead == NULL) {
            assert(curRxLength == -1);
            assert(!rxDone);

            curRxHead = check_and_cast<HeadFlit*>(msg);

            curRxLength = 1;
        } else {
            assert(curRxLength >= 1);
            assert(!rxDone);
            
            curRxLength++;
            
            TailFlit *tail = dynamic_cast<TailFlit*>(msg);
            if (tail != NULL) {
                rxDone = true;
            }
            
            delete msg;
        }
    }
}

}
