#include "RoutingLogic.h"

#include <assert.h>
#include <string.h>
#include <stdio.h>

#include "Util.h"
#include "Messages_m.h"

namespace YARC {

/*
 * This module represents the routing logic of a tile. It keeps track of the congestion
 * state of the current row. When an incoming head flit arrives the routing function is
 * called to produce an output port number. The number is translated to output column 
 * and the head and further flits of the current packet are sent to this column. While
 * the current output column is not congested the RoutingLogic sends an Ack message back
 * to the queues. Ack means "you may send more". If congestion happens the RoutingLogic
 * delays sending the Ack until the current output column is no longer congested. 
 */

Define_Module(RoutingLogic);

void RoutingLogic::initialize() {
    in = gate("in");
    out = gate("out");
    ack = gate("ack");
    throttle = gate("throttle");
    
    colCount = par("colCount");
    routingParams = par("customRoutingParams");
    
    cModule *switchModule = this;
    switchModule = switchModule->getParentModule();
    switchModule = switchModule->getParentModule();    
    // assert(strcmp(switchModule->getComponentType()->getName(), "Switch") == 0);    
    switchId = switchModule->getId();

    currentDstPort = -1;    
    currentVC = -1;
    pendingAckEmit = false;

    isCongested = new char[colCount][2];
    memset(isCongested, 0, sizeof(char[2]) * colCount);
    
    WATCH(switchId);
    WATCH(routingParams);    
    WATCH(currentDstPort);
    WATCH(currentVC);
    WATCH(pendingAckEmit);
    WATCH(colCount);
    for (int i=0; i < colCount; i++) {
        const int BUFF_SIZE = 100;
        char *buff = newBuff(BUFF_SIZE);      
        
        snprintf(buff, BUFF_SIZE, "isCongested[%d][0]", i);
        watchFlag(buff, isCongested[i][0]);
        
        snprintf(buff, BUFF_SIZE, "isCongested[%d][0]", i);
        watchFlag(buff, isCongested[i][1]);
    }
}

void RoutingLogic::handleMessage(cMessage *msg) {
    if (msg->getArrivalGate() == in) {
        Flit *flit = check_and_cast<Flit*>(msg);
        
        // Undo hack from TileInputQueues
        flit->setSchedulingPriority(0);

        HeadFlit *headFlit = dynamic_cast<HeadFlit*>(msg);
        if (headFlit != NULL) {
            assert(currentDstPort == -1);
            assert(currentVC == -1);
            assert(!pendingAckEmit);
            
            currentVC = headFlit->getVC();
            
            RoutingFuncPtr func = getRoutingFunc();
            currentDstPort = func(headFlit, switchId, routingParams);
            
            addLongPar(headFlit, OUTPUT_PORT_ATTR, currentDstPort);
        } 
        assert(currentDstPort != -1);
        assert(currentVC == 0 || currentVC == 1);

        int col = currentDstPort % colCount;
        assert(col >= 0 && col < colCount);
        addLongPar(flit, UNICAST_DST_ATTR, col);
        send(flit, out);
        pendingAckEmit = true;

        TailFlit *tailFlit = dynamic_cast<TailFlit*>(msg);
        if (tailFlit != NULL || !isCongested[col][currentVC]) {
            cMessage *ackMsg = new Ack();
            send(ackMsg, ack);
            pendingAckEmit = false;
        }

        if (tailFlit != NULL) {
            currentDstPort = -1;
            currentVC = -1;
        }
    } else if (msg->getArrivalGate() == throttle) {
        Sideband *sbd = check_and_cast<Sideband*>(msg);

        int srcCol = xtrLongPar(sbd, UNICAST_SRC_ATTR);
        assert(srcCol >= 0 && srcCol < colCount);

        copySideband(sbd, isCongested[srcCol]);

        if (pendingAckEmit) {
            assert(currentDstPort >= 0);
            assert(currentVC == 0 || currentVC == 1);
            
            int col = currentDstPort % colCount;
            assert(col >= 0 && col < colCount);
            
            if (!isCongested[col][currentVC]) {        
                cMessage *ackMsg = new Ack();
                send(ackMsg, ack);
                pendingAckEmit = false;
            }
        }
        
        delete msg;
    } else {
        assert(false);
    }
}

}
