#include "RowBufferArray.h"

#include <assert.h>
#include <stdio.h>

#include "Util.h"

namespace YARC {

/*
 * This module inherits the clocked buffer array and implements the row-buffer-specific
 * flow control, ie. the congestion state of the peer column buffers is tracked with the
 * colBufferCongested array ( which is updated when a new sidebad message arrives over
 * the throttle gate ). Sending a flit is allowed if the target column buffer isn't congested.
 * The switch-internal flow control is wormhole, ie. a packet may be blocked halfway if 
 * the destination gets congested. 
 */

Define_Module(RowBufferArray);

void RowBufferArray::initialize() {
    ClockedBufferArray::initialize();
    
    throttle = gate("throttle");
    
    currentOutputRow = -1;
    
    int peerCount = getPeerCount();
    colBufferCongested = new char[peerCount][2];
    memset(colBufferCongested, 0, sizeof(char[2]) * peerCount);
    
    WATCH(currentOutputRow);
    for (int i=0; i < peerCount; i++) {
        const int BUFF_SIZE = 100;
        char *buff = newBuff(BUFF_SIZE);      
        
        snprintf(buff, BUFF_SIZE, "colBufferCongested[%d][0]", i);        
        watchFlag(buff, colBufferCongested[i][0]);
        
        snprintf(buff, BUFF_SIZE, "colBufferCongested[%d][1]", i);
        watchFlag(buff, colBufferCongested[i][1]);
    }
}

bool RowBufferArray::canSendHeadFlit(HeadFlit *head) {
    assert(currentOutputRow == -1);
    assert(getCurrentOutputVC() == -1);        
    
    int vc = head->getVC();
    assert(vc == 0 || vc == 1);
    
    if (colBufferCongested[currentOutputRow][vc]) {        
        return false;
    } else {
        int outputPort = head->par(OUTPUT_PORT_ATTR).longValue();     
        currentOutputRow = outputPort / getPeerCount();
        xtrLongPar(head, OUTPUT_PORT_ATTR);
        addLongPar(head, UNICAST_DST_ATTR, currentOutputRow);
            
        return true;
    }
}

bool RowBufferArray::canSendFlit(Flit* flit) {
    assert(currentOutputRow >= 0 && currentOutputRow < getPeerCount());
            
    int vc = getCurrentOutputVC();        
    assert(vc == 0 || vc == 1);
    
    if (colBufferCongested[currentOutputRow][vc]) {        
        return false;
    } else {
        addLongPar(flit, UNICAST_DST_ATTR, currentOutputRow);
        
        return true;
    }
}

bool RowBufferArray::canSendTailFlit(TailFlit *tail) {
    assert(currentOutputRow >= 0 && currentOutputRow < getPeerCount());
    
    int vc = getCurrentOutputVC();   
    assert(vc == 0 || vc == 1);
    
    if (colBufferCongested[currentOutputRow][vc]) {        
        return false;
    } else {
        addLongPar(tail, UNICAST_DST_ATTR, currentOutputRow);
        currentOutputRow = -1;
        
        return true;
    }
}

void RowBufferArray::handleMessage(cMessage *msg) {
    if (msg->getArrivalGate() == throttle) {
        Sideband *sbd = check_and_cast<Sideband*>(msg);
        int throttleSrc = sbd->par(UNICAST_SRC_ATTR);
        
        assert(throttleSrc >= 0 && throttleSrc <= getPeerCount());
        copySideband(sbd, colBufferCongested[throttleSrc]);
        
        delete sbd;
    } else {
        ClockedBufferArray::handleMessage(msg);
    }
}


} 
 
