#include "SidebandAdder.h"

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include "Messages_m.h"
#include "Util.h"

namespace YARC {

/*
 * This module passes flits coming from the in gate through the out gate after
 * putting the current sideband in them. The "current" sideband is the last one
 * received through the sideband gate. In the model the tile input queues are
 * connected to the sideband adder.  
 */

Define_Module(SidebandAdder);

void SidebandAdder::initialize() {
    in  = gate("in");
    out = gate("out");
    sidebandIn = gate("sideband");
    
    clockStart = par("saClockStart");
    WATCH(clockStart);
    
    clockPeriod = par("saClockPeriod");
    WATCH(clockPeriod);
    
    sideband[0] = 0;
    WATCH_FLAG(sideband[0]);
    
    sideband[1] = 0;
    WATCH_FLAG(sideband[1]);
    
    lastSentFlit = simTime();
    lastSbdUpdate = simTime();
    
    newFlit = NULL;
    
    cMessage *msg = new cMessage();
    msg->setSchedulingPriority(32000);
    scheduleAt(clockStart, msg);
}

void SidebandAdder::handleMessage(cMessage *msg) {
    if (msg->getArrivalGate() == sidebandIn) {
        // update sideband information
        
        Sideband *sbd = check_and_cast<Sideband*>(msg);
        
        if (sbd->getSideBand(0) != sideband[0] || sbd->getSideBand(1) != sideband[1]) {
            copySideband(sbd, sideband);
            lastSbdUpdate = simTime();
        }
        
        delete sbd;
    } else if (msg->getArrivalGate() == in) {
        // cache new flit
        
        Flit *flit = check_and_cast<Flit*>(msg);
        assert(newFlit == NULL);
        newFlit = flit;
    } else if (msg->isSelfMessage()) {
        delete msg;  
        
        simtime_t now = simTime();
        
        if (newFlit != NULL) {
            // inject sideband information
            copySideband(sideband, newFlit);
            lastSentFlit = now;
            send(newFlit, out);
            newFlit = NULL;
        } else if (lastSbdUpdate > lastSentFlit) {        
            IdleFlit *idle = new IdleFlit();
            
            copySideband(sideband, idle);
            
            lastSentFlit = now;
            send(idle, out);
        }
        
        cMessage *msg = new cMessage();
        msg->setSchedulingPriority(32000);
        scheduleAt(now + clockPeriod, msg);
    } else {
        assert(false);
    }
}

}

