#include "TileInputQueues.h"

#include <assert.h>

#include "Messages_m.h"
#include "Util.h"

namespace YARC {

/*
 * This module represents the pair of input queues on the tile ( or switch port ).
 * It signals congestion to the sideband adder through the sideband gate. It receives
 * new flits over the in gate and transmits them to the routing logic over the out gate. 
 * The current virtual channel is varried in a round-robin fashion. Once a flit has been
 * sent to the routing logic no further flits are sent until an Ack message is received
 * from the routing logic.
 */

Define_Module(TileInputQueues);

void TileInputQueues::initialize() {
    in = gate("in");
    out = gate("out");
    ack = gate("ack");
    sideband = gate("sideband");
    
    threshold = par("fillThreshold");
    depth = par("queueDepth");
    
    isCongested[0] = 0;
    isCongested[1] = 0;
    
    nextOutputVC = 0;    
    currentOutputVC = -1;
    currentInputVC = -1;
    pendingAck = false;
    
    clockStart = par("queuesClockStart");
    clockPeriod = par("queuesClockPeriod");
    
    cMessage *msg = new cMessage();
    msg->setSchedulingPriority(32000);
    scheduleAt(clockStart, msg);
    
    WATCH(threshold);
    WATCH(depth);
    
    WATCH_OBJ(queues[0]);
    WATCH_FLAG(isCongested[0]);
    
    WATCH_OBJ(queues[1]);
    WATCH_FLAG(isCongested[1]);
    
    WATCH(currentInputVC);
    
    WATCH(currentOutputVC);
    WATCH(nextOutputVC);    
    WATCH(pendingAck);
}

void TileInputQueues::acceptFlit(cMessage *msg) {
    Flit *flit = check_and_cast<Flit*>(msg);

    HeadFlit *head_flit = dynamic_cast<HeadFlit*>(flit);
    if (head_flit != NULL) {
        assert(currentInputVC == -1);
        currentInputVC = head_flit->getVC();
    }

    assert(currentInputVC == 0 || currentInputVC == 1);
    queues[currentInputVC].insert(flit);

    TailFlit *tail_flit = dynamic_cast<TailFlit*>(flit);
    if (tail_flit != NULL) {
        currentInputVC = -1;
    }
    
    if (depth != -1) {
        assert(queues[0].getLength() <= depth);
        assert(queues[1].getLength() <= depth);
    }
}

void TileInputQueues::roundRobinStep() {
    currentOutputVC = nextOutputVC;
    nextOutputVC = 1 - nextOutputVC;
}

void TileInputQueues::sendFlit() {
    assert(!pendingAck);
    
    if (currentOutputVC == -1) {
        roundRobinStep();
        if (queues[currentOutputVC].isEmpty()) {
            roundRobinStep();
            if (queues[currentOutputVC].isEmpty()) {
                currentOutputVC = -1;
                return;
            }
        }
    }
    
    if (queues[currentOutputVC].isEmpty()) {
        return;
    }
    
    pendingAck = true;
            
    assert(currentOutputVC == 0 || currentOutputVC == 1);
    cObject *obj = queues[currentOutputVC].pop();
    Flit *flit = check_and_cast<Flit*>(obj);    
    
    TailFlit *tailFlit = dynamic_cast<TailFlit*>(flit);
    if (tailFlit != NULL) {
        currentOutputVC = -1;
    }
    
    // Hack to make the routing logic process throttles first without clocking it explicitly
    flit->setSchedulingPriority(1000);
    
    send(flit, out);
}

void TileInputQueues::updateSideband() {
    char currentSideband[2] = {0, 0};
    bool mustUpdate = false;
    
    for (int vc=0; vc < 2; vc++) {
        if (queues[vc].getLength() >= threshold) {
            currentSideband[vc] = 1;
        }
        if (currentSideband[vc] != isCongested[vc]) {
            mustUpdate = true;
        }
    }
    
    if (mustUpdate) {
        Sideband *sbd = new Sideband();
        copySideband(currentSideband, sbd);
        copySideband(currentSideband, isCongested);
        send(sbd, sideband);
    }
}

void TileInputQueues::handleMessage(cMessage *msg) {
    if (msg->getArrivalGate() == in) {
        acceptFlit(msg);
    } else if (msg->getArrivalGate() == ack) {
        assert(pendingAck);
        delete msg;
        pendingAck = false;
    } else if (msg->isSelfMessage()) {
        delete msg;
        
        if (!pendingAck) {
            sendFlit();
        }        
        updateSideband();
        
        cMessage *msg = new cMessage();
        msg->setSchedulingPriority(32000);
        simtime_t nextClock = simTime();
        nextClock += clockPeriod;
        scheduleAt(nextClock, msg);
    } else {
        assert(false);
    }
}

}
